import json
from time import time

from scrapy.utils.engine import get_engine_status
from twisted.web import resource
from scrapy.utils.trackref import format_live_refs, NoneType, live_refs


class JsonResource(resource.Resource):
    json_encoder = json.JSONEncoder()

    def render(self, txrequest):
        r = resource.Resource.render(self, txrequest)
        return self.render_object(r, txrequest)

    def render_object(self, obj, txrequest):
        # 获取引擎状态
        ext = format_engine_status_test(obj.engine)
        # 获取各个排列状态
        refs = format_live_refs_test()
        # 获取所有统计信息：
        stats = obj.stats.get_stats()
        # fix encode error
        stats_str = {}
        for k, v in stats.items():
            stats_str[k.decode()] = v.decode()
        list_data = [ext, refs, stats_str]
        r = self.json_encoder.encode(list_data)
        sb = bytes(r, encoding="utf8")
        txrequest.setHeader('Content-Type', 'application/json')
        txrequest.setHeader('Access-Control-Allow-Origin', '*')
        txrequest.setHeader('Access-Control-Allow-Methods', 'GET, POST, PATCH, PUT, DELETE')
        txrequest.setHeader('Access-Control-Allow-Headers', 'X-Requested-With')
        txrequest.setHeader('Content-Length', str(len(sb)))
        return sb

def format_live_refs_test(ignore=NoneType):
    now = time()
    map = {}
    for cls, wdict in sorted(live_refs.items(),
                             key=lambda x: x[0].__name__):
        if not wdict:
            continue
        if issubclass(cls, ignore):
            continue
        oldest = min(wdict.values())
        s = f"{len(wdict)} oldest: {int(now - oldest)}s ago"
        map[cls.__name__] = s
    return map


def format_engine_status_test(engine=None):
    checks = get_engine_status(engine)
    map = {}
    for test, result in checks:
        map[test] = result

    return map



import logging

from twisted.web import server, resource

from scrapy.exceptions import NotConfigured
from scrapy import signals
from scrapy.utils.reactor import listen_tcp

from scrapy_jsonrpc.jsonrpc import jsonrpc_server_call
from scrapy_jsonrpc.serialize import ScrapyJSONEncoder, ScrapyJSONDecoder
from scrapy_jsonrpc.txweb import JsonResource as JsonResource_

logger = logging.getLogger(__name__)


class JsonResource(JsonResource_):

    def __init__(self, crawler, target=None):
        JsonResource_.__init__(self)
        self.crawler = crawler
        self.json_encoder = ScrapyJSONEncoder(crawler=crawler)


class JsonRpcResource(JsonResource):

    def __init__(self, crawler, target=None):
        JsonResource.__init__(self, crawler, target)
        self.json_decoder = ScrapyJSONDecoder(crawler=crawler)
        self.crawler = crawler
        self._target = target

    def render_GET(self, txrequest):
        return self.get_target()

    def render_POST(self, txrequest):
        reqstr = txrequest.content.getvalue()
        target = self.get_target()
        return jsonrpc_server_call(target, reqstr, self.json_decoder)

    def getChild(self, name, txrequest):
        target = self.get_target()
        try:
            newtarget = getattr(target, name)
            return JsonRpcResource(self.crawler, newtarget)
        except AttributeError:
            return resource.ErrorPage(404, "No Such Resource", "No such child resource.")

    def get_target(self):
        return self._target


class CrawlerResource(JsonRpcResource):
    ws_name = 'crawler'

    def __init__(self, crawler):
        JsonRpcResource.__init__(self, crawler, crawler)


class RootResource(JsonResource):

    def render_GET(self, txrequest):
        return {'resources': self.children.keys()}

    def getChild(self, name, request):
        if name == '':
            return self
        return JsonResource.getChild(self, name, request)


class WebService(server.Site):

    def __init__(self, crawler):
        if not crawler.settings.getbool('JSONRPC_ENABLED'):
            raise NotConfigured
        self.crawler = crawler
        logfile = crawler.settings['JSONRPC_LOGFILE']
        self.portrange = [int(x) for x in crawler.settings.getlist('JSONRPC_PORT', [6023, 6073])]
        self.host = crawler.settings.get('JSONRPC_HOST', '127.0.0.1')
        root = RootResource(crawler)
        root.putChild(b'crawler', CrawlerResource(self.crawler))
        server.Site.__init__(self, root, logPath=logfile)
        self.noisy = False
        crawler.signals.connect(self.start_listening, signals.engine_started)
        crawler.signals.connect(self.stop_listening, signals.engine_stopped)

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler)

    def start_listening(self):
        self.port = listen_tcp(self.portrange, self.host, self)
        logger.info(
            'Web service listening on {host.host:s}:{host.port:d}'.format(
                host=self.port.getHost()))

    def stop_listening(self):
        self.port.stopListening()



from os.path import dirname, join
from setuptools import setup, find_packages


setup(
    name='scrapy-jsonrpc-api-py3',
    version='1.0.2',
    url='https://github.com/songhao8080/scrapy-jsonrpc',
    description='Scrapy extenstion to control spiders using JSON-RPC By songhao8080',
    author='Scrapy developers,songhao8080',
    license='BSD',
    packages=find_packages(exclude=('tests', 'tests.*')),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Framework :: Scrapy',
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
    ],
    install_requires=[
        'Twisted>=10.0.0',
        'Scrapy>=2.4.0',
        'six>=1.15.0',
        'w3lib>=1.22.0',
        'urllib3>=1.26.2',
    ],
)
